/* Shortest path in directed graph with integer weights > 0 */

#include <stdio.hs>
#include "cmtime.hs"

/* Maximum number of nodes */
#define N 1500

/* Maximum number of edges */
#define E 6500

/* Upper bound on path length (for demonstration) */
#define INF 999999999

/* Graph representation */

domain node { int d, p, x, i;  domain node *q; };

/* d - Distance from start, INF if not yet reached 
   p - Predecessor in current path from start 
   x - status: -1 - not used, 0 - no path yet, 1 - candidate,
		2 - newly solved, 4 - solved
   i - relative position in domain 
   q - pointer to corresponding node in reversal */

domain edge { domain node *f, *t;  int lng; };

/* f, t - Edge is from f to t
   lng  - Length of edge */

domain node ND[N];  domain edge EG[E];

int NN, NE;  /* Actual number of nodes, edges */

FILE *graphp;     /* Stream for input file */

void loadg()
 {int n, t, d;
  NN = NE = 0;
  [domain node].{ x = -1;  i = (this - &ND[0]); };
  [domain edge].{ f = t = NULL;  d = -1; };
  fscanf(graphp, "%d", &n);
  while ( n >= 0 )
   { if ( ND[n].x == -1 ) { ND[n].x = 0;  NN++; };
     fscanf(graphp, "%d%d", &t, &d);
     while ( t >= 0 )
      { EG[NE].f = &ND[n];  EG[NE].t = &ND[t];  EG[NE].lng = d;  NE++;
        fscanf(graphp, "%d%d", &t, &d);
      };
     fscanf(graphp, "%d", &n);
    };
 }

/* Input graph description :

 0  t00 d00  t01 d01 ... t0n d0n  -1 -1  {edges from 0}
 1  t10 d10  t11 d11 ... t1n d1n  -1 -1  {edges from 1}
      .
      .  {-1's are stoppers, nodes may have any out-degrees >= 0}
      .
 m  tm0 dm0  tm1 dm1 ... tmn dmn  -1 -1
-1  {node stopper}
{'from' nodes don't have to be in order, may be repeated}
*/

void printg()
 {int n, j, k;
  printf(
" fr   to lng   to lng   to lng   to lng   to lng   to lng   to lng\n");
  [domain node].
   { int act;
     act = (x >= 0);
     while (act)
      { n = (<?= i);  ND[n].act = 0;  
        printf("\n%3d", n);  k = 7;
        [domain edge].
         { int live;
           live = ( f == &ND[n] ) ? (this - &EG[0]) : -1;
           while (live >= 0)
            { j = (<?= live);
              printf("  %3d %3d", EG[j].t->i, EG[j].lng);
	      EG[j].live = -1;
              k++;  if (k%7 == 0) { printf("\n   ");  k = 0; };
            };
	 };
        if ( k != 0 ) printf("\n");
       };
   }
 }


/* Print paths.  Assumes active nodes are 0, ..., NN-1. */

void printpath()
 { int i, j, jx;
   for ( i = 0; i < NN; i += 15 )
    { jx = (NN - i) <? 15;
      printf("\ni");
      for ( j = 0; j < jx; j++) printf(" %4d", i+j);
      printf("\np");
      for ( j = 0; j < jx; j++) printf(" %4d", ND[i+j].p);
      printf("\nd");
      for ( j = 0; j < jx; j++) printf(" %4d", (ND[i+j].d <? 9999));
      printf("\nx");
      for ( j = 0; j < jx; j++) printf(" %4d", ND[i+j].x);
      printf("\n");
    }
 }

/* Array for multiple source calculation */

int NS, SRC[N];

/* Find shortest paths from source(s) to all other nodes
   s >= 0 - single source s
   s <  0 - NS sources from SRC
   Display codes:  0 - none, 1 - printpaths */

int path(s, dsply)
 int s, dsply;
 {int w, more, eps, k;
  [domain node].{ p = -1;  d = INF;  if ( x >= 0 ) x = 0; };
  [domain edge].{ if ( f != NULL ) eps = (<?= lng); };
  if ( s >= 0 ) { ND[s].d = 0;  ND[s].x = 1; }
     else for ( k = 0; k < NS; k++ )
              { s = SRC[k];  ND[s].d = 0;  ND[s].x = 1; };
  more = 1;  k = 0;
  while ( more )
     { if ( dsply == 1) printpath();  k++;
       [domain node].
	{ if ( x == 1 ) { w = (<?= d);
		          if ( d <= w+eps ) x = 2;
		        };
	};
       [domain edge].
	{ if ( f != NULL )
	   { int dx, tx;
	     if ( f->x == 2 )         /* edge from new */
		{ dx = f->d + lng;    /* distance to t */
		  tx = t->x;
 		  if ( tx <= 1 )      /* update distances */ 
		   { t->d <?= dx;
		     if ( tx == 0)  t->x = 1;
		     if ( t->d == dx ) t->p ,= f->i ; }
		};
	   };
 	 };
        [domain node].{ if ( x == 2 ) x = 4;  more = (|= (x == 1)); };
      };
  return(k);
 }

/* Double the size of the current graph.  Incoporates edges of length
   cl in both directions for first and last nodes.  Assumes nodes are
   0..NN-1. */

void doubleg(cl)
 int cl;
 {int nx, ex;
  nx = 2*NN;  ex = 2*NE;
  if ( (nx > N) || (ex + 4 > E) )
	{ printf("Graph is too big to double.\n"); return; };
  [domain node].{ if ( i < nx ) x = 0; };  /* Copy nodes */
  [domain edge].			   /* Copy edges */
   { domain edge *p;
     p = this + NE;
     if ( f != NULL )
	{ p->f = f + NN;  p->t = t + NN;  p->lng = lng; };
   };
  /* Insert connecting edges */
  EG[ex].f = &ND[0];     EG[ex].t = &ND[NN];    EG[ex].lng = cl;  ex++;
  EG[ex].f = &ND[NN];    EG[ex].t = &ND[0];     EG[ex].lng = cl;  ex++;
  EG[ex].f = &ND[NN-1];  EG[ex].t = &ND[nx-1];  EG[ex].lng = cl;  ex++;
  EG[ex].f = &ND[nx-1];  EG[ex].t = &ND[NN-1];  EG[ex].lng = cl;  ex++;
  NN = nx;  NE = ex;
  printf("The graph now has %3d nodes and %4d edges.\n", NN, NE);
 }

/* Program for bi-directional search from src to tgt.
   Status codes for back search are 8 + forward codes.
   See path for display codes */

int bipath(src, tgt, dsply)
 int src, tgt, dsply;
 {int wf, wb, more, eps, k, nx, ex, met;
  /* Construct reverse graph and initialize */
  nx = 2*NN;  ex = 2*NE;
  if ( (nx > N) || (ex > E) )
	{ printf("Graph is too big to reverse.\n"); return(0); };
  [domain node].          /* Copy nodes */
   { if ( i < nx )
      { if ( i < NN ) { x = 0;  q = this + NN; } else x = 8; };
      p = -1;  d = INF;
   };
  [domain edge].			   /* Copy edges */
   { domain edge *q;
     if ( f != NULL )
	{ q = this+NE;  q->t = f+NN;  q->f = t+NN;  q->lng = lng;
	  eps = (<?= lng);
	};
   };
  if ( dsply >= 1 ) printf("epsilon = %d\n", eps);
  if ( dsply >= 2 ) printg();    /* DEBUG */
  ND[src].x = 1;  ND[tgt+NN].x = 9;  ND[tgt+NN].d = ND[src].d = 0;
  more = 1;  met = 0;  k = 0;  NN = nx;
  while ( more && !met )
     { if ( dsply >= 1) printpath();  k++;
       [domain node].
	{ if ( x == 1 ) { wf = (<?= d);  if ( d <= wf+eps ) x =  2; };
	  if ( x == 9 ) { wb = (<?= d);  if ( d <= wb+eps ) x = 10; };
	};
       if ( dsply >= 1 ) printf("wf = %d wb = %d\n", wf, wb);
       [domain edge].
	{ if ( f != NULL )
	   { int dx, tx;
	     if ( (f->x)%8 == 2 )     /* edge from new candidate */
		{ dx = f->d + lng;    /* distance to t */
		  tx = (t->x)%8;
 		  if ( tx <= 1 )      /* update distances, paths */ 
		   { t->d <?= dx;
		     if ( tx == 0)  t->x |= 1;
		     if ( t->d == dx ) t->p ,= f->i ; }
		};
	   };
 	 };
        [domain node].{ if ( x%8 == 2 ) x += 2;
			more = (|= (x == 1)) & (|= (x == 9));
			if ( x == 4 ) met = (|= (q->x == 12));
		      };
        if ( dsply >= 1 ) printf("more = %d  met = %d\n", more, met);
      };
  if ( dsply >= 1 ) printpath();
  NN >>= 1;      /* Original value */
  if ( met )     /* Determine optimal path */
  {int dmin, link, lf, lb, rf, rb;
   [domain node].
    { int dt;
      if ( (x == 4) && (q->x >= 9) ) { dt = d + q->d;  dmin = (<?= dt); };
      if ( dt == dmin ) link ,= i;   /* Link node for opt. path */
    };
   for ( lf = link, lb = lf+NN;  lf != tgt;  lf = rf, lb = rb )
    { rb = ND[lb].p;  rf = rb-NN;  ND[rf].p = lf;  ND[rf].x = 4;
      ND[rf].d = ND[lf].d + (ND[lb].d - ND[rb].d);
    };
  };
  /* Restore graph */
  [domain node].{ if ( i >= NN ) p = x = -1; };
  [domain edge].
   { if ( (f != NULL) && (f->i >= NN) ) { f = t = NULL;  lng = -1; }};
  return(k);
 }

/* Test program.  Commands are
  l <file>      -  load graph from named file
  g             -  print graph
  d <int>       -  set display status to int
                   -1 - none, 0 - result, 1 - intermediate
  p <start>     -  find paths from start and print
  m <src>... -1 -  find paths from multiple sources, -1 to end list
  b <src> <tgt> -  bi-directional search from src to tgt
  2 <int>       -  double size of graph, connecting edges of lng int
  q	        -  quit
*/

void main()
 { char c[2], fnm[50];  int start, dest, display, x;
   long timer;
   display = 0;
   while ( printf("path>"), scanf("%1s", c), c[0] != 'q' )
    switch (c[0])
     { case 'l':  scanf("%s", fnm);  graphp = fopen(fnm, "r");
		  loadg();  fclose(graphp);
		  printf("loaded %3d nodes, %3d edges\n", NN, NE);
		  break;
       case 'g':  printg();
		  break;
       case 'd':  scanf("%d", &display);  break;
       case 'p':  scanf("%d", &start);
		  timer = clock();  x = path(start, display);
		  timer = clock() - timer;
		  if ( display >= 0 ) printpath();
		  printf("\n%d iterations, %ld microsec.\n", x, timer);
		  break;
       case 'b':  scanf("%d%d", &start, &dest);
		  timer = clock();  x = bipath(start, dest, display);
		  timer = clock() - timer;
		  if ( display >= 0 ) printpath();
		  printf("\n%d iterations, %ld microsec.\n", x, timer);
		  break;
       case 'm':  NS = 0;
		  while ( scanf("%d", &start), start >= 0 )
		   { SRC[NS] = start;  NS++; };
		  timer = clock();  x = path(-1, display);
		  timer = clock() - timer;
		  if ( display >= 0 ) printpath();
		  printf("\n%d iterations, %ld microsec.\n", x, timer);
		  break;
       case '2':  scanf("%d", &x);  doubleg(x);  break;
       default:   printf("Bad input - ignored\n");  break;
     };
 }
